/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: EventService.java 4389 2008-12-15 13:48:57Z alitokmen $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.event.impl;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.ow2.util.event.api.IEventDispatcher;
import org.ow2.util.event.api.IEventListener;
import org.ow2.util.event.api.IEventService;

/**
 * Allow to register event dispatchers with an id.<br>
 * Each listeners are associate with a regexp used to register the listener on
 * all matching dispatchers.
 * @author missonng
 */
public class EventService implements IEventService {
    /**
     * The event service state.
     */
    private boolean available = false;

    /**
     * Map ids to dispatcher.
     */
    private Map<String, IEventDispatcher> ids2dispatchers = new HashMap<String, IEventDispatcher>();

    /**
     * Map listeners to regexps.
     */
    private Map<IEventListener, String> listeners2regexps = new HashMap<IEventListener, String>();

    /**
     * Map dispatchers to ids to listeners.
     */
    private Map<IEventDispatcher, Map<String, Set<IEventListener>>> dispatchers2ids2listeners =
        new HashMap<IEventDispatcher, Map<String, Set<IEventListener>>>();

    /**
     * Map dispatchers to listeners to ids.
     */
    private Map<IEventDispatcher, Map<IEventListener, Set<String>>> dispatchers2listeners2ids =
        new HashMap<IEventDispatcher, Map<IEventListener, Set<String>>>();

    /**
     * Start the service.
     */
    public synchronized void start() {
        if (!this.available) {
            this.available = true;

            for (Entry<IEventDispatcher, Map<IEventListener, Set<String>>> entry : this.dispatchers2listeners2ids.entrySet()) {
                final IEventDispatcher dispatcher = entry.getKey();
                final Set<IEventListener> listeners = entry.getValue().keySet();

                for (IEventListener listener : listeners) {
                    dispatcher.addListener(listener);
                }
            }
        }
    }

    /**
     * Stop the service.
     */
    public synchronized void stop() {
        if (this.available) {
            this.available = false;

            for (Entry<IEventDispatcher, Map<IEventListener, Set<String>>> entry : this.dispatchers2listeners2ids.entrySet()) {
                final IEventDispatcher dispatcher = entry.getKey();
                final Set<IEventListener> listeners = entry.getValue().keySet();

                for (IEventListener listener : listeners) {
                    dispatcher.removeListener(listener);
                }
            }
        }
    }

    /**
     * Get the event dispatcher with the given id.
     * @param id The id of the dispatcher to get.
     * @return The event dispatcher with the given id.
     */
    public synchronized IEventDispatcher getDispatcher(final String id) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        return this.ids2dispatchers.get(id);
    }

    /**
     * Register a new event dispatcher.
     * @param id The id to register the dispatcher with.
     * @param dispatcher The dispatcher to register.
     */
    public synchronized void registerDispatcher(final String id, final IEventDispatcher dispatcher) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        if (id == null) {
            throw new IllegalArgumentException();
        }

        unregisterDispatcher(id);

        this.ids2dispatchers.put(id, dispatcher);

        Map<IEventListener, Set<String>> listeners2ids = this.dispatchers2listeners2ids.get(dispatcher);
        if (listeners2ids == null) {
            listeners2ids = new HashMap<IEventListener, Set<String>>();
            this.dispatchers2listeners2ids.put(dispatcher, listeners2ids);
        }

        Map<String, Set<IEventListener>> ids2listeners = this.dispatchers2ids2listeners.get(dispatcher);
        if (ids2listeners == null) {
            ids2listeners = new HashMap<String, Set<IEventListener>>();
            this.dispatchers2ids2listeners.put(dispatcher, ids2listeners);
        }

        final Set<IEventListener> listeners = new HashSet<IEventListener>();
        ids2listeners.put(id, listeners);

        for (Entry<IEventListener, String> entry : this.listeners2regexps.entrySet()) {
            final IEventListener listener = entry.getKey();
            final String regexp = entry.getValue();

            if (id.matches(regexp)) {
                Set<String> ids = listeners2ids.get(listener);

                if (ids == null) {
                    ids = new HashSet<String>();
                    listeners2ids.put(listener, ids);
                    dispatcher.addListener(listener);
                }

                ids.add(id);
                listeners.add(listener);
            }
        }
    }

    /**
     * Unregister an event dispatcher.
     * @param id The id to unregister.
     */
    public synchronized void unregisterDispatcher(final String id) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        if (id == null) {
            throw new IllegalArgumentException();
        }

        final IEventDispatcher dispatcher = this.ids2dispatchers.remove(id);

        if (dispatcher != null) {
            final Map<String, Set<IEventListener>> ids2listeners = this.dispatchers2ids2listeners.get(dispatcher);
            final Map<IEventListener, Set<String>> listeners2ids = this.dispatchers2listeners2ids.get(dispatcher);

            final Set<IEventListener> listeners = ids2listeners.remove(id);
            if (ids2listeners.size() == 0) {
                this.dispatchers2ids2listeners.remove(dispatcher);
                this.dispatchers2listeners2ids.remove(dispatcher);
            }

            for (IEventListener listener : listeners) {
                final Set<String> ids = listeners2ids.get(listener);

                ids.remove(id);
                if (ids.size() == 0) {
                    listeners2ids.remove(listener);
                    dispatcher.removeListener(listener);
                }
            }
        }
    }

    /**
     * Register a new event listener.<br>
     * The listener will automatically be register with each registered
     * dispatcher matching the given regexp.
     * @param listener The listener to register.
     * @param regexp The regexp to apply.
     */
    public synchronized void registerListener(final IEventListener listener, final String regexp) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        unregisterListener(listener);

        this.listeners2regexps.put(listener, regexp);

        for (Entry<String, IEventDispatcher> entry : this.ids2dispatchers.entrySet()) {
            final String id = entry.getKey();
            final IEventDispatcher dispatcher = entry.getValue();

            if (id.matches(regexp)) {
                final Map<String, Set<IEventListener>> ids2listeners = this.dispatchers2ids2listeners.get(dispatcher);
                final Map<IEventListener, Set<String>> listeners2ids = this.dispatchers2listeners2ids.get(dispatcher);

                Set<IEventListener> listeners = ids2listeners.get(id);
                Set<String> ids = listeners2ids.get(listener);

                if (ids == null) {
                    ids = new HashSet<String>();
                    listeners2ids.put(listener, ids);
                    dispatcher.addListener(listener);
                }

                ids.add(id);
                listeners.add(listener);
            }
        }
    }

    /**
     * Unregister an event listener.
     * @param listener The listener to unregister.
     */
    public synchronized void unregisterListener(final IEventListener listener) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        final String regexp = this.listeners2regexps.remove(listener);

        if (regexp != null) {
            for (Entry<String, IEventDispatcher> entry : this.ids2dispatchers.entrySet()) {
                final String id = entry.getKey();
                final IEventDispatcher dispatcher = entry.getValue();

                if (id.matches(regexp)) {
                    final Map<String, Set<IEventListener>> ids2listeners = this.dispatchers2ids2listeners.get(dispatcher);
                    final Map<IEventListener, Set<String>> listeners2ids = this.dispatchers2listeners2ids.get(dispatcher);
                    final Set<IEventListener> listeners = ids2listeners.get(id);
                    final Set<String> ids = listeners2ids.get(listener);

                    ids.remove(id);
                    listeners.remove(listener);

                    if (ids.size() == 0) {
                        listeners2ids.remove(listener);
                        dispatcher.removeListener(listener);
                    }
                }
            }
        }
    }
}